﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Optimization.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.VehicleRouting.Encodings.Alba;
using HeuristicLab.Problems.VehicleRouting.Encodings.Potvin;

namespace HeuristicLab.Problems.VehicleRouting {
  /// <summary>
  /// An operator that improves vehicle routing solutions.
  /// </summary>
  [Item("VRPImprovementOperator", "An operator that improves vehicle routing solutions.")]
  [StorableClass]
  public sealed class VRPImprovementOperator : SingleSuccessorOperator, IImprovementOperator {
    #region Parameter properties
    public ILookupParameter<DoubleValue> CapacityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Capacity"]; }
    }
    public ILookupParameter<DoubleMatrix> CoordinatesParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["Coordinates"]; }
    }
    public ScopeParameter CurrentScopeParameter {
      get { return (ScopeParameter)Parameters["CurrentScope"]; }
    }
    public ILookupParameter<DoubleArray> DemandParameter {
      get { return (ILookupParameter<DoubleArray>)Parameters["Demand"]; }
    }
    public ILookupParameter<DoubleValue> EvalDistanceFactorParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["EvalDistanceFactor"]; }
    }
    public ILookupParameter<DoubleMatrix> DistanceMatrixParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["DistanceMatrix"]; }
    }
    public ILookupParameter<DoubleArray> DueTimeParameter {
      get { return (ILookupParameter<DoubleArray>)Parameters["DueTime"]; }
    }
    public ILookupParameter<IEvaluator> EvaluatorParameter {
      get { return (ILookupParameter<IEvaluator>)Parameters["Evaluator"]; }
    }
    public ILookupParameter<DoubleValue> EvalFleetUsageFactorParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["EvalFleetUsageFactor"]; }
    }
    public IValueParameter<IntValue> ImprovementAttemptsParameter {
      get { return (IValueParameter<IntValue>)Parameters["ImprovementAttempts"]; }
    }
    public IValueParameter<IntValue> LambdaParameter {
      get { return (IValueParameter<IntValue>)Parameters["Lambda"]; }
    }
    public ILookupParameter<DoubleValue> EvalOverloadPenaltyParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["EvalOverloadPenalty"]; }
    }
    public ILookupParameter<IRandom> RandomParameter {
      get { return (ILookupParameter<IRandom>)Parameters["Random"]; }
    }
    public ILookupParameter<DoubleArray> ReadyTimeParameter {
      get { return (ILookupParameter<DoubleArray>)Parameters["ReadyTime"]; }
    }
    public IValueParameter<IntValue> SampleSizeParameter {
      get { return (IValueParameter<IntValue>)Parameters["SampleSize"]; }
    }
    public ILookupParameter<DoubleArray> ServiceTimeParameter {
      get { return (ILookupParameter<DoubleArray>)Parameters["ServiceTime"]; }
    }
    public ILookupParameter<DoubleValue> EvalTardinessPenaltyParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["EvalTardinessPenalty"]; }
    }
    public IValueLookupParameter<IItem> TargetParameter {
      get { return (IValueLookupParameter<IItem>)Parameters["Target"]; }
    }
    public ILookupParameter<DoubleValue> EvalTimeFactorParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["EvalTimeFactor"]; }
    }
    public ILookupParameter<BoolValue> UseDistanceMatrixParameter {
      get { return (ILookupParameter<BoolValue>)Parameters["UseDistanceMatrix"]; }
    }
    public ILookupParameter<IntValue> VehiclesParameter {
      get { return (ILookupParameter<IntValue>)Parameters["Vehicles"]; }
    }
    #endregion

    [StorableConstructor]
    private VRPImprovementOperator(bool deserializing) : base(deserializing) { }
    private VRPImprovementOperator(VRPImprovementOperator original, Cloner cloner) : base(original, cloner) { }
    public VRPImprovementOperator()
      : base() {
      #region Create parameters
      Parameters.Add(new LookupParameter<DoubleValue>("Capacity"));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Coordinates"));
      Parameters.Add(new ScopeParameter("CurrentScope"));
      Parameters.Add(new LookupParameter<DoubleArray>("Demand"));
      Parameters.Add(new LookupParameter<DoubleValue>("EvalDistanceFactor"));
      Parameters.Add(new LookupParameter<DoubleMatrix>("DistanceMatrix"));
      Parameters.Add(new LookupParameter<DoubleArray>("DueTime"));
      Parameters.Add(new LookupParameter<IEvaluator>("Evaluator"));
      Parameters.Add(new LookupParameter<DoubleValue>("EvalFleetUsageFactor"));
      Parameters.Add(new ValueParameter<IntValue>("ImprovementAttempts", new IntValue(100)));
      Parameters.Add(new ValueParameter<IntValue>("Lambda", new IntValue(1)));
      Parameters.Add(new LookupParameter<DoubleValue>("EvalOverloadPenalty"));
      Parameters.Add(new LookupParameter<IRandom>("Random"));
      Parameters.Add(new LookupParameter<DoubleArray>("ReadyTime"));
      Parameters.Add(new ValueParameter<IntValue>("SampleSize", new IntValue(100)));
      Parameters.Add(new LookupParameter<DoubleArray>("ServiceTime"));
      Parameters.Add(new LookupParameter<DoubleValue>("EvalTardinessPenalty"));
      Parameters.Add(new ValueLookupParameter<IItem>("Target"));
      Parameters.Add(new LookupParameter<DoubleValue>("EvalTimeFactor"));
      Parameters.Add(new LookupParameter<BoolValue>("UseDistanceMatrix"));
      Parameters.Add(new LookupParameter<IntValue>("Vehicles"));
      #endregion
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new VRPImprovementOperator(this, cloner);
    }

    public override IOperation Apply() {
      AlbaEncoding solution = TargetParameter.ActualValue is AlbaEncoding ?
        TargetParameter.ActualValue as AlbaEncoding :
        AlbaEncoding.ConvertFrom(TargetParameter.ActualValue as IVRPEncoding, VehiclesParameter.ActualValue.Value, DistanceMatrixParameter);

      double quality = -1;
      int evaluatedSolutions;

      AlbaLambdaInterchangeLocalImprovementOperator.Apply(solution, ImprovementAttemptsParameter.Value.Value, true, LambdaParameter.Value.Value,
        SampleSizeParameter.Value.Value, RandomParameter.ActualValue, ref quality, out evaluatedSolutions, DistanceMatrixParameter.ActualValue,
        VehiclesParameter.ActualValue, DueTimeParameter.ActualValue, ServiceTimeParameter.ActualValue, ReadyTimeParameter.ActualValue,
        DemandParameter.ActualValue, CapacityParameter.ActualValue, CoordinatesParameter.ActualValue, EvalFleetUsageFactorParameter.ActualValue,
        EvalTimeFactorParameter.ActualValue, EvalDistanceFactorParameter.ActualValue, EvalOverloadPenaltyParameter.ActualValue,
        EvalTardinessPenaltyParameter.ActualValue, UseDistanceMatrixParameter.ActualValue);

      TargetParameter.ActualValue = PotvinEncoding.ConvertFrom(solution, DistanceMatrixParameter);

      return base.Apply();
    }
  }
}
